from typing import Dict, Any

import orjson
from tortoise import fields, Model
from tortoise.indexes import Index

from eventual.infra import mixin
from eventual.infra.exchange.abc import ProcessingGuarantee


class PkUuidModel(Model):
    id = fields.UUIDField(pk=True)

    class Meta:
        abstract = True


def _dump_str(mapping: Dict[str, Any]) -> str:
    return orjson.dumps(mapping).decode()


class EventOutRelation(Model, mixin.Timestamp):
    event_id = fields.UUIDField()
    body = fields.JSONField(encoder=_dump_str, decoder=orjson.loads)
    confirmed = fields.BooleanField(default=False)
    send_after = fields.DatetimeField(index=True, null=True, default=None)

    class Meta:
        table = "event_out"
        indexes = [
            Index(fields={"created_at"}),
        ]


class DispatchedEventRelation(Model, mixin.Timestamp):
    body = fields.JSONField(encoder=_dump_str, decoder=orjson.loads)
    event_id = fields.UUIDField()

    class Meta:
        table = "dispatched_event"


class HandledEventRelation(PkUuidModel, mixin.Timestamp):
    body = fields.JSONField(encoder=_dump_str, decoder=orjson.loads)
    guarantee = fields.CharEnumField(ProcessingGuarantee)

    class Meta:
        table = "handled_event"



import abc
import uuid
from typing import TypeVar, Iterable, Optional, Type, Generic

from eventual.repo import (
    RepositoryAllMixin,
    RepositoryCountMixin,
    RepositoryGetMixin,
    RepositoryCreateMixin,
    RepositoryUpdateMixin,
    RepositoryDeleteMixin,
)
from eventual.infra.relation import PkUuidModel
from eventual.model import Entity

M = TypeVar("M", bound=Entity)
R = TypeVar("R", bound=PkUuidModel)


class TortoiseRepository(Generic[M, R], abc.ABC):
    @property
    @abc.abstractmethod
    def _relation_cls(self) -> Type[R]:
        raise NotImplementedError

    @abc.abstractmethod
    def store(self, obj: M) -> R:
        raise NotImplementedError

    @abc.abstractmethod
    async def reconstitute(self, obj: R) -> M:
        raise NotImplementedError


class TortoiseRepositoryAllMixin(
    RepositoryAllMixin[M], TortoiseRepository[M, R], abc.ABC
):
    async def all(self) -> Iterable[M]:
        seq = await self._relation_cls.all()
        return [await self.reconstitute(p) for p in seq]


class TortoiseRepositoryCountMixin(
    RepositoryCountMixin[M], TortoiseRepository[M, R], abc.ABC
):
    async def count(self) -> int:
        return await self._relation_cls.all().count()


class TortoiseRepositoryGetMixin(
    RepositoryGetMixin[M], TortoiseRepository[M, R], abc.ABC
):
    async def get(self, unique_id: uuid.UUID) -> Optional[M]:
        p = await self._relation_cls.filter(id=unique_id).get_or_none()
        if p is None:
            return None
        return await self.reconstitute(p)


class TortoiseRepositoryCreateMixin(
    RepositoryCreateMixin[M], TortoiseRepository[M, R], abc.ABC
):
    async def create(self, value: M) -> M:
        p = self.store(value)
        await p.save()
        return await self.reconstitute(p)


class TortoiseRepositoryUpdateMixin(
    RepositoryUpdateMixin[M], TortoiseRepository[M, R], abc.ABC
):
    async def update(self, value: M) -> Optional[M]:
        p = await self._relation_cls.filter(id=value.id).get_or_none()
        if p is None:
            return None
        p = self.store(value)
        await p.save()
        return await self.reconstitute(p)


class TortoiseRepositoryDeleteMixin(
    RepositoryDeleteMixin[M], TortoiseRepository[M, R], abc.ABC
):
    async def delete_by_id(self, unique_id: uuid.UUID) -> Optional[M]:
        p = await self._relation_cls.filter(id=unique_id).get_or_none()
        if p is None:
            return None
        await self._relation_cls.filter(id=unique_id).delete()
        return await self.reconstitute(p)


class TortoiseRepositoryMixin(
    TortoiseRepositoryAllMixin[M, R],
    TortoiseRepositoryCountMixin[M, R],
    TortoiseRepositoryGetMixin[M, R],
    TortoiseRepositoryCreateMixin[M, R],
    TortoiseRepositoryUpdateMixin[M, R],
    TortoiseRepositoryDeleteMixin[M, R],
    abc.ABC,
):
    pass



from typing import AsyncIterable, AsyncGenerator, Optional, Callable

import aio_pika
import orjson

from eventual.infra.exchange.abc import Message, MessageExchange, EventBody


class AmqpMessage(Message):
    def __init__(self, message: aio_pika.IncomingMessage):
        self._body = orjson.loads(message.body)
        self._message = message

    @property
    def body(self) -> EventBody:
        return self._body

    def acknowledge(self):
        self._message.ack()


class AmqpMessageExchange(MessageExchange):
    def __init__(
        self,
        amqp_dsn: str,
        amqp_exchange: str,
        amqp_queue: str,
        routing_key_from_type: Optional[Callable[[str], str]] = None,
    ):
        self.amqp_dsn = amqp_dsn
        self.amqp_exchange = amqp_exchange
        self.amqp_queue = amqp_queue

        def default_routing_key_from_type(event_type: str) -> str:
            return f"{self.amqp_queue}.{event_type}"

        self.routing_key_from_type = default_routing_key_from_type

    async def _message_stream(self, connection) -> AsyncGenerator[AmqpMessage, None]:
        channel: aio_pika.Channel = await connection.channel()
        async with channel:
            exchange = await channel.declare_exchange(
                self.amqp_exchange, aio_pika.ExchangeType.FANOUT
            )
            queue: aio_pika.Queue = await channel.declare_queue(
                self.amqp_queue, durable=True
            )
            await queue.bind(exchange)

            async with queue.iterator() as queue_iter:
                async for message in queue_iter:
                    yield AmqpMessage(message)

    async def message_stream(self) -> AsyncIterable[AmqpMessage]:
        connection: aio_pika.RobustConnection = await aio_pika.connect_robust(
            self.amqp_dsn
        )

        async with connection:
            return self._message_stream(connection)

    async def send_event_body_stream(self, event_body_stream: AsyncIterable[EventBody]):
        connection: aio_pika.RobustConnection = await aio_pika.connect_robust(
            self.amqp_dsn
        )

        async with connection:
            channel: aio_pika.Channel = await connection.channel(
                publisher_confirms=True
            )
            exchange = await channel.declare_exchange(
                self.amqp_exchange, aio_pika.ExchangeType.FANOUT
            )

            async for event_body in event_body_stream:
                event_type = event_body["type"]
                await exchange.publish(
                    aio_pika.Message(
                        body=orjson.dumps(event_body),
                        delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
                    ),
                    routing_key=self.routing_key_from_type(event_type),
                )



import contextlib
from typing import AsyncGenerator

from tortoise.transactions import in_transaction

from .abc import UnitOfWork, WorkInterrupted


class TortoiseUnitOfWork(UnitOfWork):
    async def rollback(self):
        raise WorkInterrupted


@contextlib.asynccontextmanager
async def tortoise_unit_of_work() -> AsyncGenerator[TortoiseUnitOfWork, None]:
    try:
        async with in_transaction("default"):
            uow = TortoiseUnitOfWork()
            yield uow
            uow.was_committed = True
    except WorkInterrupted:
        pass



import asyncio
import uuid
from contextlib import asynccontextmanager
from typing import AsyncGenerator

from eventual import util
from eventual.infra.exchange.abc import (
    EventStorage,
    EventBody,
    Message,
    ProcessingGuarantee,
)
from eventual.infra.relation import (
    EventOutRelation,
    HandledEventRelation,
    DispatchedEventRelation,
)
from eventual.infra.uow import tortoise_unit_of_work
from eventual.model import Entity


class RelationalEventStorage(EventStorage):
    @asynccontextmanager
    async def clear_outbox_atomically(
        self, *entity_seq: Entity
    ) -> AsyncGenerator[None, None]:
        async with tortoise_unit_of_work():
            yield None
            await self.clear_outbox(entity_seq)

    async def schedule_event_out(
        self, event_id: uuid.UUID, body: EventBody, send_after: bool = False
    ):
        await EventOutRelation.create(
            event_id=event_id, body=body, send_after=send_after
        )

    async def is_event_handled(self, event_id: uuid.UUID) -> bool:
        event_count = await HandledEventRelation.filter(id=event_id).count()
        return event_count > 0

    async def mark_event_handled(
        self, event_body: EventBody, guarantee: ProcessingGuarantee
    ) -> uuid.UUID:
        event_id = event_body["id"]
        await HandledEventRelation.create(
            id=event_id, body=event_body, guarantee=guarantee
        )
        return event_id

    async def mark_event_dispatched(self, event_body: EventBody) -> uuid.UUID:
        event_id = event_body["id"]
        await DispatchedEventRelation.create(body=event_body, event_id=event_id)
        return event_id

    @asynccontextmanager
    async def handle_exactly_once(
        self, message: Message
    ) -> AsyncGenerator[EventBody, None]:
        async with tortoise_unit_of_work():
            yield message.body
            await self.mark_event_handled(
                message.body, guarantee=ProcessingGuarantee.EXACTLY_ONCE
            )
        message.acknowledge()

    @asynccontextmanager
    async def handle_no_more_than_once(
        self, message: Message
    ) -> AsyncGenerator[EventBody, None]:
        await self.mark_event_handled(
            message.body, guarantee=ProcessingGuarantee.NO_MORE_THAN_ONCE
        )
        message.acknowledge()
        yield message.body

    @asynccontextmanager
    async def handle_at_least_once(
        self, message: Message
    ) -> AsyncGenerator[EventBody, None]:
        yield message.body
        await self.mark_event_handled(
            message.body, guarantee=ProcessingGuarantee.AT_LEAST_ONCE
        )
        message.acknowledge()

    async def event_body_stream(self) -> AsyncGenerator[EventBody, None]:
        while True:
            event_seq = await EventOutRelation.filter(
                confirmed=False, send_after__lt=util.tz_aware_utcnow()
            ).order_by("created_at")

            for event in event_seq:
                yield event.body
                event.confirmed = True
                await event.save()
            await asyncio.sleep(1)



import abc
from typing import Callable, AsyncContextManager


class WorkInterrupted(Exception):
    pass


class UnitOfWork(abc.ABC):
    def __init__(self):
        self.was_committed = False

    @abc.abstractmethod
    async def rollback(self):
        raise NotImplementedError


UnitOfWorkFactory = Callable[[], AsyncContextManager[UnitOfWork]]



import uuid
from typing import TypeVar, Iterable, Optional, Generic, Dict

from eventual.repo import (
    RepositoryAllMixin,
    RepositoryCountMixin,
    RepositoryGetMixin,
    RepositoryCreateMixin,
    RepositoryUpdateMixin,
    RepositoryDeleteMixin,
)
from eventual.model import Entity

M = TypeVar("M", bound=Entity)


class MemoryRepository(Generic[M]):
    def __init__(self):
        self._mapping: Dict[uuid.UUID, M] = {}


class MemoryRepositoryAllMixin(RepositoryAllMixin[M], MemoryRepository[M]):
    async def all(self) -> Iterable[M]:
        return list(self._mapping.values())


class MemoryRepositoryCountMixin(RepositoryCountMixin[M], MemoryRepository[M]):
    async def count(self) -> int:
        return len(self._mapping)


class MemoryRepositoryGetMixin(RepositoryGetMixin[M], MemoryRepository[M]):
    async def get(self, unique_id: uuid.UUID) -> Optional[M]:
        return self._mapping.get(unique_id)


class MemoryRepositoryCreateMixin(RepositoryCreateMixin[M], MemoryRepository[M]):
    async def create(self, value: M) -> M:
        self._mapping[value.id] = value
        return value


class MemoryRepositoryUpdateMixin(RepositoryUpdateMixin[M], MemoryRepository[M]):
    async def update(self, value: M) -> Optional[M]:
        m = self._mapping.get(value.id)
        if m is None:
            return None
        self._mapping[value.id] = value
        return value


class MemoryRepositoryDeleteMixin(RepositoryDeleteMixin[M], MemoryRepository[M]):
    async def delete_by_id(self, unique_id: uuid.UUID) -> Optional[M]:
        m = self._mapping.get(unique_id)
        if m is None:
            return None
        return self._mapping.pop(unique_id)


class MemoryRepositoryMixin(
    MemoryRepositoryAllMixin[M],
    MemoryRepositoryCountMixin[M],
    MemoryRepositoryGetMixin[M],
    MemoryRepositoryCreateMixin[M],
    MemoryRepositoryUpdateMixin[M],
    MemoryRepositoryDeleteMixin[M],
):
    pass



import asyncio
import datetime as dt
import uuid
from typing import AsyncIterable, Awaitable, Callable, Dict, List

from eventual import util
from eventual.infra.exchange.abc import (
    EventStorage,
    Message,
    MessageDispatcher,
    MessageExchange,
)


async def _wait_and_pop(
    message_dispatcher: "ConcurrentMessageDispatcher",
    unique_task_id: uuid.UUID,
    fn: Callable[[Message], Awaitable[None]],
    message: Message,
):
    try:
        await fn(message)
    except Exception as e:
        send_after = util.tz_aware_utcnow() + dt.timedelta(
            seconds=message_dispatcher.delay_on_exc
        )
        body = message.body
        await message_dispatcher.event_storage.schedule_event_out(
            body["id"],
            body,
            send_after=send_after,
        )
        # TODO: Make sure that no as few messages as possible get stuck in a sent/rejected loop.
        # This should not violate any guarantees, but can possibly create a lot of messages.
        message.acknowledge()
        print(e)
    finally:
        _ = message_dispatcher.task_from_id.pop(unique_task_id)


class ConcurrentMessageDispatcher(MessageDispatcher):
    def __init__(self, event_storage: EventStorage, delay_on_exc: float):
        self.event_storage = event_storage
        self.delay_on_exc = delay_on_exc

        self.fn_from_message_type: Dict[str, Callable[[Message], Awaitable[None]]] = {}
        self.task_from_id: Dict[uuid.UUID, asyncio.Task] = {}

    def register(
        self, event_type_seq: List[str], fn: Callable[[Message], Awaitable[None]]
    ):
        for event_type in event_type_seq:
            if event_type in self.fn_from_message_type:
                # TODO: Change error type to something more appropriate.
                raise ValueError(
                    "it is not possible to register multiple functions to handle the same event type"
                )
            self.fn_from_message_type[event_type] = fn

    async def dispatch_from_exchange(self, message_exchange: MessageExchange):
        while True:
            message_stream = await message_exchange.message_stream()
            self.interrupt_dispatch = False
            await self.dispatch_from_stream(message_stream)

    async def dispatch_from_stream(self, message_stream: AsyncIterable[Message]):
        try:
            async for message in message_stream:
                event_body = message.body

                is_event_handled = await self.event_storage.is_event_handled(
                    event_body["id"]
                )
                if is_event_handled:
                    # There is no guarantee that messages that we've marked as handled
                    # were actually acknowledged, so it's not an error to get the handled message.
                    # Furthermore, even if someone sends the same message multiple times,
                    # but we consider it handled, we do nothing in a truly idempotent manner.
                    message.acknowledge()
                    continue

                if (
                    fn := self.fn_from_message_type.get(event_body["type"])
                ) is not None:
                    # We save every event that we attempt to dispatch, not every event we receive,
                    # because we can get a lot of events that we do not care about.
                    await self.event_storage.mark_event_dispatched(event_body)

                    # We want to remove tasks from `self.task_from_id` upon their completion.
                    # To do that we have to associate each one with a unique key. Such a key
                    # should be unique for every task, so we can't use a message id -
                    # during the handling of a message we can get the same message again.
                    # For now we've settled on a uuid, but it's certainly not the fastest option.
                    unique_id = uuid.uuid4()
                    task = asyncio.create_task(
                        _wait_and_pop(
                            self,
                            unique_id,
                            fn,
                            message,
                        )
                    )
                    self.task_from_id[unique_id] = task
        except asyncio.CancelledError as e:
            pass
        except Exception as e:
            # TODO: Add logging here to signal that we are shutting down.
            raise
        finally:
            await asyncio.gather(*self.task_from_id.values(), return_exceptions=True)



import contextlib
from typing import AsyncGenerator

from .abc import UnitOfWork, WorkInterrupted


class MemoryUnitOfWork(UnitOfWork):
    def __init__(self):
        super().__init__()

    async def rollback(self):
        raise WorkInterrupted


@contextlib.asynccontextmanager
async def memory_unit_of_work() -> AsyncGenerator[MemoryUnitOfWork, None]:
    try:
        uow = MemoryUnitOfWork()
        yield uow
        uow.was_committed = True
    except WorkInterrupted:
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eventual',
 'eventual.infra',
 'eventual.infra.exchange',
 'eventual.infra.repo',
 'eventual.infra.uow',
 'eventual.model',
 'eventual.repo',
 'eventual.util']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=6.8.0,<7.0.0',
 'orjson>=3.5.2,<4.0.0',
 'tortoise-orm[asyncpg]>=0.17.2,<0.18.0']

setup_kwargs = {
    'name': 'eventual',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': 'Ivan Dmitriesvkii',
    'author_email': 'ivan.dmitrievsky@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import asyncio

from eventual.infra.exchange.abc import (
    EventStorage,
    MessageDispatcher,
    MessageExchange,
)


def register_gum(
    app,
    message_exchange: MessageExchange,
    message_dispatcher: MessageDispatcher,
    event_storage: EventStorage,
):
    task_from_name = dict()

    @app.on_event("startup")
    async def start_gum():
        task_from_name["dispatch_task"] = asyncio.create_task(
            message_dispatcher.dispatch_from_exchange(message_exchange)
        )
        task_from_name["publish_task"] = asyncio.create_task(
            message_exchange.send_event_body_stream(event_storage.event_body_stream())
        )

    @app.on_event("shutdown")
    async def stop_gum():
        # This can be called with a delay.
        # For example, uvicorn apparently waits for all tasks to complete,
        # before it invokes this callback. The timeout seems to be 10 seconds,
        # and we don't know if it can be configured.
        # Nothing actually changes for us, because nothing gets cancelled.
        task_from_name["dispatch_task"].cancel()
        task_from_name["publish_task"].cancel()
        await asyncio.gather(*task_from_name.values(), return_exceptions=True)
